@@ -21,6 +21,7 @@ template <class S, class T> inline void clone(T*& dst, S* src, int n)
 	memcpy((void *)dst,(void *)src,sizeof(T)*n);
 }
 #define INF HUGE_VAL
+#define TAU 1e-12
 #define Malloc(type,n) (type *)malloc((n)*sizeof(type))
 #if 1
 void info(char *fmt,...)
@@ -66,7 +67,7 @@ private:
 		int len;		// data[0,len) is cached in this entry
 	};
 
-	head_t* head;
+	head_t *head;
 	head_t lru_head;
 	void lru_delete(head_t *h);
 	void lru_insert(head_t *h);
@@ -77,6 +78,7 @@ Cache::Cache(int l_,int size_):l(l_),size(size_)
 	head = (head_t *)calloc(l,sizeof(head_t));	// initialized to 0
 	size /= sizeof(Qfloat);
 	size -= l * sizeof(head_t) / sizeof(Qfloat);
+	size = max(size, 2*l);	// cache must be large enough for two columns
 	lru_head.next = lru_head.prev = &lru_head;
 }
 
@@ -171,7 +173,14 @@ void Cache::swap_index(int i, int j)
 // the constructor of Kernel prepares to calculate the l*l kernel matrix
 // the member function get_Q is for getting one column from the Q Matrix
 //
-class Kernel {
+class QMatrix {
+public:
+	virtual Qfloat *get_Q(int column, int len) const = 0;
+	virtual Qfloat *get_QD() const = 0;
+	virtual void swap_index(int i, int j) const = 0;
+};
+
+class Kernel: public QMatrix {
 public:
 	Kernel(int l, svm_node * const * x, const svm_parameter& param);
 	virtual ~Kernel();
@@ -179,6 +188,7 @@ public:
 	static double k_function(const svm_node *x, const svm_node *y,
 				 const svm_parameter& param);
 	virtual Qfloat *get_Q(int column, int len) const = 0;
+	virtual Qfloat *get_QD() const = 0;
 	virtual void swap_index(int i, int j) const	// no so const...
 	{
 		swap(x[i],x[j]);
@@ -365,7 +375,7 @@ public:
 		double r;	// for Solver_NU
 	};
 
-	void Solve(int l, const Kernel& Q, const double *b_, const schar *y_,
+	void Solve(int l, const QMatrix& Q, const double *b_, const schar *y_,
 		   double *alpha_, double Cp, double Cn, double eps,
 		   SolutionInfo* si, int shrinking);
 protected:
@@ -375,7 +385,8 @@ protected:
 	enum { LOWER_BOUND, UPPER_BOUND, FREE };
 	char *alpha_status;	// LOWER_BOUND, UPPER_BOUND, FREE
 	double *alpha;
-	const Kernel *Q;
+	const QMatrix *Q;
+	const Qfloat *QD;
 	double eps;
 	double Cp,Cn;
 	double *b;
@@ -402,6 +413,7 @@ protected:
 	void swap_index(int i, int j);
 	void reconstruct_gradient();
 	virtual int select_working_set(int &i, int &j);
+	virtual int max_violating_pair(int &i, int &j);
 	virtual double calculate_rho();
 	virtual void do_shrinking();
 };
@@ -438,12 +450,13 @@ void Solver::reconstruct_gradient()
 		}
 }
 
-void Solver::Solve(int l, const Kernel& Q, const double *b_, const schar *y_,
+void Solver::Solve(int l, const QMatrix& Q, const double *b_, const schar *y_,
 		   double *alpha_, double Cp, double Cn, double eps,
 		   SolutionInfo* si, int shrinking)
 {
 	this->l = l;
 	this->Q = &Q;
+	QD=Q.get_QD();
 	clone(b, b_,l);
 	clone(y, y_,l);
 	clone(alpha,alpha_,l);
@@ -480,7 +493,7 @@ void Solver::Solve(int l, const Kernel& Q, const double *b_, const schar *y_,
 		for(i=0;i<l;i++)
 			if(!is_lower_bound(i))
 			{
-				Qfloat *Q_i = Q.get_Q(i,l);
+				const Qfloat *Q_i = Q.get_Q(i,l);
 				double alpha_i = alpha[i];
 				int j;
 				for(j=0;j<l;j++)
@@ -536,7 +549,10 @@ void Solver::Solve(int l, const Kernel& Q, const double *b_, const schar *y_,
 
 		if(y[i]!=y[j])
 		{
-			double delta = (-G[i]-G[j])/max(Q_i[i]+Q_j[j]+2*Q_i[j],(Qfloat)0);
+			double quad_coef = Q_i[i]+Q_j[j]+2*Q_i[j];
+			if (quad_coef <= 0)
+				quad_coef = TAU;
+			double delta = (-G[i]-G[j])/quad_coef;
 			double diff = alpha[i] - alpha[j];
 			alpha[i] += delta;
 			alpha[j] += delta;
@@ -576,10 +592,14 @@ void Solver::Solve(int l, const Kernel& Q, const double *b_, const schar *y_,
 		}
 		else
 		{
-			double delta = (G[i]-G[j])/max(Q_i[i]+Q_j[j]-2*Q_i[j],(Qfloat)0);
+			double quad_coef = Q_i[i]+Q_j[j]-2*Q_i[j];
+			if (quad_coef <= 0)
+				quad_coef = TAU;
+			double delta = (G[i]-G[j])/quad_coef;
 			double sum = alpha[i] + alpha[j];
 			alpha[i] -= delta;
 			alpha[j] += delta;
+
 			if(sum > C_i)
 			{
 				if(alpha[i] > C_i)
@@ -701,14 +721,107 @@ void Solver::Solve(int l, const Kernel& Q, const double *b_, const schar *y_,
 // return 1 if already optimal, return 0 otherwise
 int Solver::select_working_set(int &out_i, int &out_j)
 {
-	// return i,j which maximize -grad(f)^T d , under constraint
-	// if alpha_i == C, d != +1
-	// if alpha_i == 0, d != -1
+	// return i,j such that
+	// i: maximizes -y_i * grad(f)_i, i in I_up(\alpha)
+	// j: minimizes the decrease of obj value
+	//    (if quadratic coefficeint <= 0, replace it with tau)
+	//    -y_j*grad(f)_j < -y_i*grad(f)_i, j in I_low(\alpha)
+	
+	double Gmax = -INF;
+	int Gmax_idx = -1;
+	int Gmin_idx = -1;
+	double obj_diff_min = INF;
 
-	double Gmax1 = -INF;		// max { -grad(f)_i * d | y_i*d = +1 }
+	for(int t=0;t<active_size;t++)
+		if(y[t]==+1)	
+		{
+			if(!is_upper_bound(t))
+				if(-G[t] >= Gmax)
+				{
+					Gmax = -G[t];
+					Gmax_idx = t;
+				}
+		}
+		else
+		{
+			if(!is_lower_bound(t))
+				if(G[t] >= Gmax)
+				{
+					Gmax = G[t];
+					Gmax_idx = t;
+				}
+		}
+
+	int i = Gmax_idx;
+	const Qfloat *Q_i = NULL;
+	if(i != -1) // NULL Q_i not accessed: Gmax=-INF if i=-1
+		Q_i = Q->get_Q(i,active_size);
+
+	for(int j=0;j<active_size;j++)
+	{
+		if(y[j]==+1)
+		{
+			if (!is_lower_bound(j))
+			{
+				double grad_diff=Gmax+G[j];
+				if (grad_diff >= eps)
+				{
+					double obj_diff; 
+					double quad_coef=Q_i[i]+QD[j]-2*y[i]*Q_i[j];
+					if (quad_coef > 0)
+						obj_diff = -(grad_diff*grad_diff)/quad_coef;
+					else
+						obj_diff = -(grad_diff*grad_diff)/TAU;
+
+					if (obj_diff <= obj_diff_min)
+					{
+						Gmin_idx=j;
+						obj_diff_min = obj_diff;
+					}
+				}
+			}
+		}
+		else
+		{
+			if (!is_upper_bound(j))
+			{
+				double grad_diff= Gmax-G[j];
+				if (grad_diff >= eps)
+				{
+					double obj_diff; 
+					double quad_coef=Q_i[i]+QD[j]+2*y[i]*Q_i[j];
+					if (quad_coef > 0)
+						obj_diff = -(grad_diff*grad_diff)/quad_coef;
+					else
+						obj_diff = -(grad_diff*grad_diff)/TAU;
+
+					if (obj_diff <= obj_diff_min)
+					{
+						Gmin_idx=j;
+						obj_diff_min = obj_diff;
+					}
+				}
+			}
+		}
+	}
+
+	if(Gmin_idx == -1)
+ 		return 1;
+
+	out_i = Gmax_idx;
+	out_j = Gmin_idx;
+	return 0;
+}
+
+// return 1 if already optimal, return 0 otherwise
+int Solver::max_violating_pair(int &out_i, int &out_j)
+{
+	// return i,j: maximal violating pair
+
+	double Gmax1 = -INF;		// max { -y_i * grad(f)_i | i in I_up(\alpha) }
 	int Gmax1_idx = -1;
 
-	double Gmax2 = -INF;		// max { -grad(f)_i * d | y_i*d = -1 }
+	double Gmax2 = -INF;		// max { y_i * grad(f)_i | i in I_low(\alpha) }
 	int Gmax2_idx = -1;
 
 	for(int i=0;i<active_size;i++)
@@ -717,7 +830,7 @@ int Solver::select_working_set(int &out_i, int &out_j)
 		{
 			if(!is_upper_bound(i))	// d = +1
 			{
-				if(-G[i] > Gmax1)
+				if(-G[i] >= Gmax1)
 				{
 					Gmax1 = -G[i];
 					Gmax1_idx = i;
@@ -725,7 +838,7 @@ int Solver::select_working_set(int &out_i, int &out_j)
 			}
 			if(!is_lower_bound(i))	// d = -1
 			{
-				if(G[i] > Gmax2)
+				if(G[i] >= Gmax2)
 				{
 					Gmax2 = G[i];
 					Gmax2_idx = i;
@@ -736,7 +849,7 @@ int Solver::select_working_set(int &out_i, int &out_j)
 		{
 			if(!is_upper_bound(i))	// d = +1
 			{
-				if(-G[i] > Gmax2)
+				if(-G[i] >= Gmax2)
 				{
 					Gmax2 = -G[i];
 					Gmax2_idx = i;
@@ -744,7 +857,7 @@ int Solver::select_working_set(int &out_i, int &out_j)
 			}
 			if(!is_lower_bound(i))	// d = -1
 			{
-				if(G[i] > Gmax1)
+				if(G[i] >= Gmax1)
 				{
 					Gmax1 = G[i];
 					Gmax1_idx = i;
@@ -764,7 +877,7 @@ int Solver::select_working_set(int &out_i, int &out_j)
 void Solver::do_shrinking()
 {
 	int i,j,k;
-	if(select_working_set(i,j)!=0) return;
+	if(max_violating_pair(i,j)!=0) return;
 	double Gm1 = -y[j]*G[j];
 	double Gm2 = y[i]*G[i];
 
@@ -875,7 +988,7 @@ class Solver_NU : public Solver
 {
 public:
 	Solver_NU() {}
-	void Solve(int l, const Kernel& Q, const double *b, const schar *y,
+	void Solve(int l, const QMatrix& Q, const double *b, const schar *y,
 		   double *alpha, double Cp, double Cn, double eps,
 		   SolutionInfo* si, int shrinking)
 	{
@@ -889,89 +1002,121 @@ private:
 	void do_shrinking();
 };
 
+// return 1 if already optimal, return 0 otherwise
 int Solver_NU::select_working_set(int &out_i, int &out_j)
 {
-	// return i,j which maximize -grad(f)^T d , under constraint
-	// if alpha_i == C, d != +1
-	// if alpha_i == 0, d != -1
+	// return i,j such that y_i = y_j and
+	// i: maximizes -y_i * grad(f)_i, i in I_up(\alpha)
+	// j: minimizes the decrease of obj value
+	//    (if quadratic coefficeint <= 0, replace it with tau)
+	//    -y_j*grad(f)_j < -y_i*grad(f)_i, j in I_low(\alpha)
 
-	double Gmax1 = -INF;	// max { -grad(f)_i * d | y_i = +1, d = +1 }
-	int Gmax1_idx = -1;
+	double Gmaxp = -INF;
+	int Gmaxp_idx = -1;
 
-	double Gmax2 = -INF;	// max { -grad(f)_i * d | y_i = +1, d = -1 }
-	int Gmax2_idx = -1;
+	double Gmaxn = -INF;
+	int Gmaxn_idx = -1;
 
-	double Gmax3 = -INF;	// max { -grad(f)_i * d | y_i = -1, d = +1 }
-	int Gmax3_idx = -1;
+	int Gmin_idx = -1;
+	double obj_diff_min = INF;
 
-	double Gmax4 = -INF;	// max { -grad(f)_i * d | y_i = -1, d = -1 }
-	int Gmax4_idx = -1;
-
-	for(int i=0;i<active_size;i++)
-	{
-		if(y[i]==+1)	// y == +1
+	for(int t=0;t<active_size;t++)
+		if(y[t]==+1)
 		{
-			if(!is_upper_bound(i))	// d = +1
-			{
-				if(-G[i] > Gmax1)
+			if(!is_upper_bound(t))
+				if(-G[t] >= Gmaxp)
 				{
-					Gmax1 = -G[i];
-					Gmax1_idx = i;
+					Gmaxp = -G[t];
+					Gmaxp_idx = t;
 				}
-			}
-			if(!is_lower_bound(i))	// d = -1
-			{
-				if(G[i] > Gmax2)
+		}
+		else
+		{
+			if(!is_lower_bound(t))
+				if(G[t] >= Gmaxn)
 				{
-					Gmax2 = G[i];
-					Gmax2_idx = i;
+					Gmaxn = G[t];
+					Gmaxn_idx = t;
+				}
+		}
+
+	int ip = Gmaxp_idx;
+	int in = Gmaxn_idx;
+	const Qfloat *Q_ip = NULL;
+	const Qfloat *Q_in = NULL;
+	if(ip != -1) // NULL Q_ip not accessed: Gmaxp=-INF if ip=-1
+		Q_ip = Q->get_Q(ip,active_size);
+	if(in != -1)
+		Q_in = Q->get_Q(in,active_size);
+
+	for(int j=0;j<active_size;j++)
+	{
+		if(y[j]==+1)
+		{
+			if (!is_lower_bound(j))	
+			{
+				double grad_diff=Gmaxp+G[j];
+				if (grad_diff >= eps)
+				{
+					double obj_diff; 
+					double quad_coef = Q_ip[ip]+QD[j]-2*Q_ip[j];
+					if (quad_coef > 0)
+						obj_diff = -(grad_diff*grad_diff)/quad_coef;
+					else
+						obj_diff = -(grad_diff*grad_diff)/TAU;
+
+					if (obj_diff <= obj_diff_min)
+					{
+						Gmin_idx=j;
+						obj_diff_min = obj_diff;
+					}
 				}
 			}
 		}
-		else		// y == -1
+		else
 		{
-			if(!is_upper_bound(i))	// d = +1
+			if (!is_upper_bound(j))
 			{
-				if(-G[i] > Gmax3)
+				double grad_diff=Gmaxn-G[j];
+				if (grad_diff >= eps)
 				{
-					Gmax3 = -G[i];
-					Gmax3_idx = i;
-				}
-			}
-			if(!is_lower_bound(i))	// d = -1
-			{
-				if(G[i] > Gmax4)
-				{
-					Gmax4 = G[i];
-					Gmax4_idx = i;
+					double obj_diff; 
+					double quad_coef = Q_in[in]+QD[j]-2*Q_in[j];
+					if (quad_coef > 0)
+						obj_diff = -(grad_diff*grad_diff)/quad_coef;
+					else
+						obj_diff = -(grad_diff*grad_diff)/TAU;
+
+					if (obj_diff <= obj_diff_min)
+					{
+						Gmin_idx=j;
+						obj_diff_min = obj_diff;
+					}
 				}
 			}
 		}
 	}
 
-	if(max(Gmax1+Gmax2,Gmax3+Gmax4) < eps)
+	if(Gmin_idx == -1)
  		return 1;
 
-	if(Gmax1+Gmax2 > Gmax3+Gmax4)
-	{
-		out_i = Gmax1_idx;
-		out_j = Gmax2_idx;
-	}
+	if (y[Gmin_idx] == +1)
+		out_i = Gmaxp_idx;
 	else
-	{
-		out_i = Gmax3_idx;
-		out_j = Gmax4_idx;
-	}
+		out_i = Gmaxn_idx;
+	out_j = Gmin_idx;
+
 	return 0;
 }
 
 void Solver_NU::do_shrinking()
 {
-	double Gmax1 = -INF;	// max { -grad(f)_i * d | y_i = +1, d = +1 }
-	double Gmax2 = -INF;	// max { -grad(f)_i * d | y_i = +1, d = -1 }
-	double Gmax3 = -INF;	// max { -grad(f)_i * d | y_i = -1, d = +1 }
-	double Gmax4 = -INF;	// max { -grad(f)_i * d | y_i = -1, d = -1 }
+	double Gmax1 = -INF;	// max { -y_i * grad(f)_i | y_i = +1, i in I_up(\alpha) }
+	double Gmax2 = -INF;	// max { y_i * grad(f)_i | y_i = +1, i in I_low(\alpha) }
+	double Gmax3 = -INF;	// max { -y_i * grad(f)_i | y_i = -1, i in I_up(\alpha) }
+	double Gmax4 = -INF;	// max { y_i * grad(f)_i | y_i = -1, i in I_low(\alpha) }
 
+	// find maximal violating pair first
 	int k;
 	for(k=0;k<active_size;k++)
 	{
@@ -993,6 +1138,8 @@ void Solver_NU::do_shrinking()
 		}
 	}
 
+	// shrinking
+
 	double Gm1 = -Gmax2;
 	double Gm2 = -Gmax1;
 	double Gm3 = -Gmax4;
@@ -1117,6 +1264,9 @@ public:
 	{
 		clone(y,y_,prob.l);
 		cache = new Cache(prob.l,(int)(param.cache_size*(1<<20)));
+		QD = new Qfloat[prob.l];
+		for(int i=0;i<prob.l;i++)
+			QD[i]= (Qfloat)(this->*kernel_function)(i,i);
 	}
 	
 	Qfloat *get_Q(int i, int len) const
@@ -1131,21 +1281,29 @@ public:
 		return data;
 	}
 
+	Qfloat *get_QD() const
+	{
+		return QD;
+	}
+
 	void swap_index(int i, int j) const
 	{
 		cache->swap_index(i,j);
 		Kernel::swap_index(i,j);
 		swap(y[i],y[j]);
+		swap(QD[i],QD[j]);
 	}
 
 	~SVC_Q()
 	{
 		delete[] y;
 		delete cache;
+		delete[] QD;
 	}
 private:
 	schar *y;
 	Cache *cache;
+	Qfloat *QD;
 };
 
 class ONE_CLASS_Q: public Kernel
@@ -1155,6 +1313,9 @@ public:
 	:Kernel(prob.l, prob.x, param)
 	{
 		cache = new Cache(prob.l,(int)(param.cache_size*(1<<20)));
+		QD = new Qfloat[prob.l];
+		for(int i=0;i<prob.l;i++)
+			QD[i]= (Qfloat)(this->*kernel_function)(i,i);
 	}
 	
 	Qfloat *get_Q(int i, int len) const
@@ -1169,18 +1330,26 @@ public:
 		return data;
 	}
 
+	Qfloat *get_QD() const
+	{
+		return QD;
+	}
+
 	void swap_index(int i, int j) const
 	{
 		cache->swap_index(i,j);
 		Kernel::swap_index(i,j);
+		swap(QD[i],QD[j]);
 	}
 
 	~ONE_CLASS_Q()
 	{
 		delete cache;
+		delete[] QD;
 	}
 private:
 	Cache *cache;
+	Qfloat *QD;
 };
 
 class SVR_Q: public Kernel
@@ -1191,6 +1360,7 @@ public:
 	{
 		l = prob.l;
 		cache = new Cache(l,(int)(param.cache_size*(1<<20)));
+		QD = new Qfloat[2*l];
 		sign = new schar[2*l];
 		index = new int[2*l];
 		for(int k=0;k<l;k++)
@@ -1199,6 +1369,8 @@ public:
 			sign[k+l] = -1;
 			index[k] = k;
 			index[k+l] = k;
+			QD[k]= (Qfloat)(this->*kernel_function)(k,k);
+			QD[k+l]=QD[k];
 		}
 		buffer[0] = new Qfloat[2*l];
 		buffer[1] = new Qfloat[2*l];
@@ -1209,6 +1381,7 @@ public:
 	{
 		swap(sign[i],sign[j]);
 		swap(index[i],index[j]);
+		swap(QD[i],QD[j]);
 	}
 	
 	Qfloat *get_Q(int i, int len) const
@@ -1230,6 +1403,11 @@ public:
 		return buf;
 	}
 
+	Qfloat *get_QD() const
+	{
+		return QD;
+	}
+
 	~SVR_Q()
 	{
 		delete cache;
@@ -1237,6 +1415,7 @@ public:
 		delete[] index;
 		delete[] buffer[0];
 		delete[] buffer[1];
+		delete[] QD;
 	}
 private:
 	int l;
@@ -1244,7 +1423,8 @@ private:
 	schar *sign;
 	int *index;
 	mutable int next_buffer;
-	Qfloat* buffer[2];
+	Qfloat *buffer[2];
+	Qfloat *QD;
 };
 
 //
@@ -1353,7 +1533,8 @@ static void solve_one_class(
 
 	for(i=0;i<n;i++)
 		alpha[i] = 1;
-	alpha[n] = param->nu * prob->l - n;
+	if(n<prob->l)
+		alpha[n] = param->nu * prob->l - n;
 	for(i=n+1;i<l;i++)
 		alpha[i] = 0;
 
@@ -1663,23 +1844,23 @@ double sigmoid_predict(double decision_value, double A, double B)
 // Method 2 from the multiclass_prob paper by Wu, Lin, and Weng
 void multiclass_probability(int k, double **r, double *p)
 {
-	int t;
+	int t,j;
 	int iter = 0, max_iter=100;
 	double **Q=Malloc(double *,k);
 	double *Qp=Malloc(double,k);
-	double pQp, eps=0.001;
+	double pQp, eps=0.005/k;
 	
 	for (t=0;t<k;t++)
 	{
 		p[t]=1.0/k;  // Valid if k = 1
 		Q[t]=Malloc(double,k);
 		Q[t][t]=0;
-		for (int j=0;j<t;j++)
+		for (j=0;j<t;j++)
 		{
 			Q[t][t]+=r[j][t]*r[j][t];
 			Q[t][j]=Q[j][t];
 		}
-		for (int j=t+1;j<k;j++)
+		for (j=t+1;j<k;j++)
 		{
 			Q[t][t]+=r[j][t]*r[j][t];
 			Q[t][j]=-r[j][t]*r[t][j];
@@ -1692,7 +1873,7 @@ void multiclass_probability(int k, double **r, double *p)
 		for (t=0;t<k;t++)
 		{
 			Qp[t]=0;
-			for (int j=0;j<k;j++)
+			for (j=0;j<k;j++)
 				Qp[t]+=Q[t][j]*p[j];
 			pQp+=p[t]*Qp[t];
 		}
@@ -1710,7 +1891,7 @@ void multiclass_probability(int k, double **r, double *p)
 			double diff=(-Qp[t]+pQp)/Q[t][t];
 			p[t]+=diff;
 			pQp=(pQp+diff*(diff*Q[t][t]+2*Qp[t]))/(1+diff)/(1+diff);
-			for (int j=0;j<k;j++)
+			for (j=0;j<k;j++)
 			{
 				Qp[j]=(Qp[j]+diff*Q[t][j])/(1+diff);
 				p[j]/=(1+diff);
@@ -1843,6 +2024,66 @@ double svm_svr_probability(
 	return mae;
 }
 
+
+// label: label name, start: begin of each class, count: #data of classes, perm: indices to the original data
+// perm, length l, must be allocated before calling this subroutine
+void svm_group_classes(const svm_problem *prob, int *nr_class_ret, int **label_ret, int **start_ret, int **count_ret, int *perm)
+{
+	int l = prob->l;
+	int max_nr_class = 16;
+	int nr_class = 0;
+	int *label = Malloc(int,max_nr_class);
+	int *count = Malloc(int,max_nr_class);
+	int *data_label = Malloc(int,l);	
+	int i;
+
+	for(i=0;i<l;i++)
+	{
+		int this_label = (int)prob->y[i];
+		int j;
+		for(j=0;j<nr_class;j++)
+		{
+			if(this_label == label[j])
+			{
+				++count[j];
+				break;
+			}
+		}
+		data_label[i] = j;
+		if(j == nr_class)
+		{
+			if(nr_class == max_nr_class)
+			{
+				max_nr_class *= 2;
+				label = (int *)realloc(label,max_nr_class*sizeof(int));
+				count = (int *)realloc(count,max_nr_class*sizeof(int));
+			}
+			label[nr_class] = this_label;
+			count[nr_class] = 1;
+			++nr_class;
+		}
+	}
+
+	int *start = Malloc(int,nr_class);
+	start[0] = 0;
+	for(i=1;i<nr_class;i++)
+		start[i] = start[i-1]+count[i-1];
+	for(i=0;i<l;i++)
+	{
+		perm[start[data_label[i]]] = i;
+		++start[data_label[i]];
+	}
+	start[0] = 0;
+	for(i=1;i<nr_class;i++)
+		start[i] = start[i-1]+count[i-1];
+
+	*nr_class_ret = nr_class;
+	*label_ret = label;
+	*start_ret = start;
+	*count_ret = count;
+	free(data_label);
+}
+
 //
 // Interface functions
 //
@@ -1896,58 +2137,19 @@ svm_model *svm_train(const svm_problem *prob, const svm_parameter *param)
 	else
 	{
 		// classification
-		// find out the number of classes
 		int l = prob->l;
-		int max_nr_class = 16;
-		int nr_class = 0;
-		int *label = Malloc(int,max_nr_class);
-		int *count = Malloc(int,max_nr_class);
-		int *index = Malloc(int,l);
-
-		int i;
-		for(i=0;i<l;i++)
-		{
-			int this_label = (int)prob->y[i];
-			int j;
-			for(j=0;j<nr_class;j++)
-				if(this_label == label[j])
-				{
-					++count[j];
-					break;
-				}
-			index[i] = j;
-			if(j == nr_class)
-			{
-				if(nr_class == max_nr_class)
-				{
-					max_nr_class *= 2;
-					label = (int *)realloc(label,max_nr_class*sizeof(int));
-					count = (int *)realloc(count,max_nr_class*sizeof(int));
-				}
-				label[nr_class] = this_label;
-				count[nr_class] = 1;
-				++nr_class;
-			}
-		}
+		int nr_class;
+		int *label = NULL;
+		int *start = NULL;
+		int *count = NULL;
+		int *perm = Malloc(int,l);
 
 		// group training data of the same class
-
-		int *start = Malloc(int,nr_class);
-		start[0] = 0;
-		for(i=1;i<nr_class;i++)
-			start[i] = start[i-1]+count[i-1];
-
+		svm_group_classes(prob,&nr_class,&label,&start,&count,perm);		
 		svm_node **x = Malloc(svm_node *,l);
-		
+		int i;
 		for(i=0;i<l;i++)
-		{
-			x[start[index[i]]] = prob->x[i];
-			++start[index[i]];
-		}
-		
-		start[0] = 0;
-		for(i=1;i<nr_class;i++)
-			start[i] = start[i-1]+count[i-1];
+			x[i] = prob->x[perm[i]];
 
 		// calculate weighted C
 
@@ -2107,7 +2309,7 @@ svm_model *svm_train(const svm_problem *prob, const svm_parameter *param)
 		free(probA);
 		free(probB);
 		free(count);
-		free(index);
+		free(perm);
 		free(start);
 		free(x);
 		free(weighted_C);
@@ -2121,26 +2323,86 @@ svm_model *svm_train(const svm_problem *prob, const svm_parameter *param)
 	return model;
 }
 
+// Stratified cross validation
 void svm_cross_validation(const svm_problem *prob, const svm_parameter *param, int nr_fold, double *target)
 {
 	int i;
-	int *perm = Malloc(int,prob->l);
+	int *fold_start = Malloc(int,nr_fold+1);
+	int l = prob->l;
+	int *perm = Malloc(int,l);
+	int nr_class;
 
-	// random shuffle
-	for(i=0;i<prob->l;i++) perm[i]=i;
-	for(i=0;i<prob->l;i++)
+	// stratified cv may not give leave-one-out rate
+	// Each class to l folds -> some folds may have zero elements
+	if((param->svm_type == C_SVC ||
+	    param->svm_type == NU_SVC) && nr_fold < l)
 	{
-		int j = i+rand()%(prob->l-i);
-		swap(perm[i],perm[j]);
+		int *start = NULL;
+		int *label = NULL;
+		int *count = NULL;
+		svm_group_classes(prob,&nr_class,&label,&start,&count,perm);
+
+		// random shuffle and then data grouped by fold using the array perm
+		int *fold_count = Malloc(int,nr_fold);
+		int c;
+		int *index = Malloc(int,l);
+		for(i=0;i<l;i++)
+			index[i]=perm[i];
+		for (c=0; c<nr_class; c++) 
+			for(i=0;i<count[c];i++)
+			{
+				int j = i+rand()%(count[c]-i);
+				swap(index[start[c]+j],index[start[c]+i]);
+			}
+		for(i=0;i<nr_fold;i++)
+		{
+			fold_count[i] = 0;
+			for (c=0; c<nr_class;c++)
+				fold_count[i]+=(i+1)*count[c]/nr_fold-i*count[c]/nr_fold;
+		}
+		fold_start[0]=0;
+		for (i=1;i<=nr_fold;i++)
+			fold_start[i] = fold_start[i-1]+fold_count[i-1];
+		for (c=0; c<nr_class;c++)
+			for(i=0;i<nr_fold;i++)
+			{
+				int begin = start[c]+i*count[c]/nr_fold;
+				int end = start[c]+(i+1)*count[c]/nr_fold;
+				for(int j=begin;j<end;j++)
+				{
+					perm[fold_start[i]] = index[j];
+					fold_start[i]++;
+				}
+			}
+		fold_start[0]=0;
+		for (i=1;i<=nr_fold;i++)
+			fold_start[i] = fold_start[i-1]+fold_count[i-1];
+		free(start);	
+		free(label);
+		free(count);	
+		free(index);
+		free(fold_count);
 	}
+	else
+	{
+		for(i=0;i<l;i++) perm[i]=i;
+		for(i=0;i<l;i++)
+		{
+			int j = i+rand()%(l-i);
+			swap(perm[i],perm[j]);
+		}
+		for(i=0;i<=nr_fold;i++)
+			fold_start[i]=i*l/nr_fold;
+	}
+
 	for(i=0;i<nr_fold;i++)
 	{
-		int begin = i*prob->l/nr_fold;
-		int end = (i+1)*prob->l/nr_fold;
+		int begin = fold_start[i];
+		int end = fold_start[i+1];
 		int j,k;
 		struct svm_problem subprob;
 
-		subprob.l = prob->l-(end-begin);
+		subprob.l = l-(end-begin);
 		subprob.x = Malloc(struct svm_node*,subprob.l);
 		subprob.y = Malloc(double,subprob.l);
 			
@@ -2151,7 +2413,7 @@ void svm_cross_validation(const svm_problem *prob, const svm_parameter *param, i
 			subprob.y[k] = prob->y[perm[j]];
 			++k;
 		}
-		for(j=end;j<prob->l;j++)
+		for(j=end;j<l;j++)
 		{
 			subprob.x[k] = prob->x[perm[j]];
 			subprob.y[k] = prob->y[perm[j]];
@@ -2173,9 +2435,11 @@ void svm_cross_validation(const svm_problem *prob, const svm_parameter *param, i
 		free(subprob.x);
 		free(subprob.y);
 	}		
+	free(fold_start);
 	free(perm);	
 }
 
+
 int svm_get_svm_type(const svm_model *model)
 {
 	return model->param.svm_type;
@@ -2597,7 +2861,8 @@ out:
 	for(i=0;i<m;i++)
 		model->sv_coef[i] = Malloc(double,l);
 	model->SV = Malloc(svm_node*,l);
-	svm_node *x_space = Malloc(svm_node,elements);
+	svm_node *x_space=NULL;
+	if(l>0) x_space = Malloc(svm_node,elements);
 
 	int j=0;
 	for(i=0;i<l;i++)
@@ -2628,7 +2893,7 @@ out2:
 
 void svm_destroy_model(svm_model* model)
 {
-	if(model->free_sv)
+	if(model->free_sv && model->l > 0)
 		free((void *)(model->SV[0]));
 	for(int i=0;i<model->nr_class-1;i++)
 		free(model->sv_coef[i]);
